import { TokenType } from "../tokenizer/types";
import * as N from "../types";
import { Pos, Position } from "../util/location";
import LValParser from "./lval";
export default abstract class ExpressionParser extends LValParser {
    abstract parseBlock(allowDirectives?: boolean, isFunctionScope?: boolean, contextId?: number): N.BlockStatement;
    abstract parseClass(node: N.Class, isStatement: boolean, optionalId?: boolean): N.Class;
    abstract parseDecorators(allowExport?: boolean): void;
    abstract parseFunction<T extends N.NormalFunction>(node: T, isStatement: boolean, allowExpressionBody?: boolean, isAsync?: boolean, optionalId?: boolean): T;
    abstract parseFunctionParams(node: N.Function, allowModifiers?: boolean, funcContextId?: number): void;
    abstract takeDecorators(node: N.HasDecorators): void;
    checkPropClash(prop: N.ObjectMember, propHash: {
        [key: string]: boolean;
    }): void;
    getExpression(): N.Expression;
    parseExpression(noIn?: boolean, refShorthandDefaultPos?: Pos): N.Expression;
    parseMaybeAssign(noIn?: boolean | null, refShorthandDefaultPos?: Pos | null, afterLeftParse?: Function, refNeedsArrowPos?: Pos | null): N.Expression;
    parseMaybeConditional(noIn: boolean | null, refShorthandDefaultPos: Pos, refNeedsArrowPos?: Pos | null): N.Expression;
    parseConditional(expr: N.Expression, noIn: boolean | null, startPos: number, startLoc: Position, refNeedsArrowPos?: Pos | null): N.Expression;
    parseExprOps(noIn: boolean | null, refShorthandDefaultPos: Pos): N.Expression;
    parseExprOp(left: N.Expression, leftStartPos: number, leftStartLoc: Position, minPrec: number, noIn: boolean | null): N.Expression;
    parseMaybeUnary(refShorthandDefaultPos?: Pos | null): N.Expression;
    parseExprSubscripts(refShorthandDefaultPos?: Pos | null): N.Expression;
    parseSubscripts(base: N.Expression, startPos: number, startLoc: Position, noCalls?: boolean | null): N.Expression;
    /** Set 'state.stop = true' to indicate that we should stop parsing subscripts. */
    parseSubscript(base: N.Expression, startPos: number, startLoc: Position, noCalls: boolean | null, state: {
        stop: boolean;
    }): N.Expression;
    atPossibleAsync(base: N.Expression): boolean;
    finishCallExpression(node: N.CallExpression): N.CallExpression;
    parseCallExpressionArguments(close: TokenType, possibleAsyncArrow: boolean, refTrailingCommaPos?: Pos): Array<N.Expression | null>;
    shouldParseAsyncArrow(): boolean;
    parseAsyncArrowFromCallExpression(node: N.ArrowFunctionExpression, call: N.CallExpression, startTokenIndex: number): N.ArrowFunctionExpression;
    parseNoCallExpr(): N.Expression;
    parseExprAtom(refShorthandDefaultPos?: Pos | null): N.Expression;
    parseBooleanLiteral(): N.BooleanLiteral;
    parseMaybePrivateName(): N.PrivateName | N.Identifier;
    parseFunctionExpression(): N.FunctionExpression | N.MetaProperty;
    parseMetaProperty(node: N.MetaProperty, meta: N.Identifier, propertyName: string): N.MetaProperty;
    parseImportMetaProperty(): N.MetaProperty;
    parseLiteral<T extends N.Literal>(value: {}, type: string, startPos?: number, startLoc?: Position): T;
    parseParenExpression(): N.Expression;
    parseParenAndDistinguishExpression(canBeArrow: boolean): N.Expression;
    shouldParseArrow(): boolean;
    parseArrow(node: N.ArrowFunctionExpression): N.ArrowFunctionExpression | null;
    parseParenItem(node: N.Expression, startPos: number, startLoc: Position): N.Expression;
    parseNew(): N.NewExpression | N.MetaProperty;
    parseNewArguments(node: N.NewExpression): void;
    parseTemplateElement(isTagged: boolean): N.TemplateElement;
    parseTemplate(isTagged: boolean): N.TemplateLiteral;
    parseObj<T extends N.ObjectPattern | N.ObjectExpression>(isPattern: boolean, isBlockScope: boolean, refShorthandDefaultPos?: Pos | null): T;
    isGetterOrSetterMethod(prop: N.ObjectMethod, isPattern: boolean): boolean;
    checkGetterSetterParamCount(method: N.ObjectMethod | N.ClassMethod): void;
    parseObjectMethod(prop: N.ObjectMethod, isGenerator: boolean, isAsync: boolean, isPattern: boolean, objectContextId: number): N.ObjectMethod | null;
    parseObjectProperty(prop: N.ObjectProperty, startPos: number | null, startLoc: Position | null, isPattern: boolean, isBlockScope: boolean, refShorthandDefaultPos: Pos | null): N.ObjectProperty | null;
    parseObjPropValue(prop: N.Node, startPos: number | null, startLoc: Position | null, isGenerator: boolean, isAsync: boolean, isPattern: boolean, isBlockScope: boolean, refShorthandDefaultPos: Pos | null, objectContextId: number): void;
    parsePropertyName(prop: N.ObjectOrClassMember | N.ClassMember | N.TsNamedTypeElementBase, objectContextId: number): N.Expression | N.Identifier;
    initFunction(node: N.BodilessFunctionOrMethodBase, isAsync: boolean): void;
    parseMethod<T extends N.MethodLike>(node: T, isGenerator: boolean, isAsync: boolean, isConstructor: boolean, type: string): T;
    parseArrowExpression(node: N.ArrowFunctionExpression, startTokenIndex: number, params?: Array<N.Expression> | null, isAsync?: boolean): N.ArrowFunctionExpression;
    setArrowFunctionParameters(node: N.ArrowFunctionExpression, params: Array<N.Expression>): void;
    isStrictBody(node: {
        body: N.BlockStatement;
    }): boolean;
    parseFunctionBodyAndFinish(node: N.BodilessFunctionOrMethodBase, type: string, allowExpressionBody?: boolean | null, funcContextId?: number): void;
    parseFunctionBody(node: N.Function, allowExpression: boolean | null, funcContextId?: number): void;
    checkFunctionNameAndParams(node: N.Function, isArrowFunction: boolean | null): void;
    parseExprList(close: TokenType, allowEmpty?: boolean | null, refShorthandDefaultPos?: Pos | null): ReadonlyArray<N.Expression | null>;
    parseExprListItem(allowEmpty: boolean | null, refShorthandDefaultPos: Pos | null, refNeedsArrowPos?: Pos | null, refTrailingCommaPos?: Pos | null): N.Expression | null;
    parseIdentifier(liberal?: boolean): N.Identifier;
    parseIdentifierName(pos: number, liberal?: boolean): string;
    checkReservedWord(word: string, startLoc: number, checkKeywords: boolean, isBinding: boolean): void;
    parseAwait(node: N.AwaitExpression): N.AwaitExpression;
    parseYield(): N.YieldExpression;
}
