import { TokenType } from "../tokenizer/types";
import { Decorator, Expression, Identifier, Node, ObjectExpression, ObjectPattern, Pattern, RestElement, SpreadElement, TSParameterProperty } from "../types";
import { Pos, Position } from "../util/location";
import { NodeUtils } from "./node";
export default abstract class LValParser extends NodeUtils {
    abstract checkReservedWord(word: string, startLoc: number, checkKeywords: boolean, isBinding: boolean): void;
    abstract parseIdentifier(liberal?: boolean): Identifier;
    abstract parseMaybeAssign(noIn?: boolean | null, refShorthandDefaultPos?: Pos | null, afterLeftParse?: Function, refNeedsArrowPos?: Pos | null): Expression;
    abstract parseObj<T extends ObjectPattern | ObjectExpression>(isPattern: boolean, isBlockScope: boolean, refShorthandDefaultPos?: Pos | null): T;
    abstract parseDecorator(): Decorator;
    toAssignable(node: Node, isBinding: boolean | null, contextDescription: string): Node;
    toAssignableObjectExpressionProp(prop: Node, isBinding: boolean | null, isLast: boolean): void;
    toAssignableList(exprList: Array<Expression>, isBinding: boolean | null, contextDescription: string): ReadonlyArray<Pattern>;
    toReferencedList(exprList: ReadonlyArray<Expression | null>): ReadonlyArray<Expression | null>;
    parseSpread<T extends RestElement | SpreadElement>(refShorthandDefaultPos: Pos | null): T;
    parseRest(isBlockScope: boolean): RestElement;
    shouldAllowYieldIdentifier(): boolean;
    parseBindingIdentifier(): Identifier;
    parseBindingAtom(isBlockScope: boolean): Pattern;
    parseBindingList(close: TokenType, isBlockScope: boolean, allowEmpty?: boolean, allowModifiers?: boolean | null): ReadonlyArray<Pattern | TSParameterProperty>;
    parseAssignableListItem(allowModifiers: boolean | null, decorators: Array<Decorator>, isBlockScope: boolean): Pattern | TSParameterProperty;
    parseAssignableListItemTypes(param: Pattern): Pattern;
    parseMaybeDefault(isBlockScope: boolean, startPos?: number | null, startLoc?: Position | null, left?: Pattern | null): Pattern;
    checkLVal(expr: Expression, isBinding: boolean | null, checkClashes: {
        [key: string]: boolean;
    } | null, contextDescription: string): void;
    checkToRestConversion(node: SpreadElement): void;
}
