import { TokenType } from "../tokenizer/types";
import * as N from "../types";
import ExpressionParser from "./expression";
export default class StatementParser extends ExpressionParser {
    parseTopLevel(file: N.File, program: N.Program): N.File;
    stmtToDirective(stmt: N.Statement): N.Directive;
    parseStatement(declaration: boolean, topLevel?: boolean): N.Statement;
    parseStatementContent(declaration: boolean, topLevel: boolean): N.Statement;
    assertModuleNodeAllowed(node: N.Node): void;
    takeDecorators(node: N.HasDecorators): void;
    parseDecorators(allowExport?: boolean): void;
    parseDecorator(): N.Decorator;
    parseBreakContinueStatement(node: N.BreakStatement | N.ContinueStatement, keyword: string): N.BreakStatement | N.ContinueStatement;
    parseDebuggerStatement(node: N.DebuggerStatement): N.DebuggerStatement;
    parseDoStatement(node: N.DoWhileStatement): N.DoWhileStatement;
    parseForStatement(node: N.Node): N.ForLike;
    parseAmbiguousForStatement(node: N.Node): N.ForLike;
    parseFunctionStatement(node: N.FunctionDeclaration): N.FunctionDeclaration;
    parseIfStatement(node: N.IfStatement): N.IfStatement;
    parseReturnStatement(node: N.ReturnStatement): N.ReturnStatement;
    parseSwitchStatement(node: N.SwitchStatement): N.SwitchStatement;
    parseThrowStatement(node: N.ThrowStatement): N.ThrowStatement;
    parseTryStatement(node: N.TryStatement): N.TryStatement;
    parseVarStatement(node: N.VariableDeclaration, kind: TokenType): N.VariableDeclaration;
    parseWhileStatement(node: N.WhileStatement): N.WhileStatement;
    parseWithStatement(node: N.WithStatement): N.WithStatement;
    parseEmptyStatement(node: N.EmptyStatement): N.EmptyStatement;
    parseLabeledStatement(node: N.LabeledStatement, maybeName: string, expr: N.Identifier): N.LabeledStatement;
    parseExpressionStatement(node: N.ExpressionStatement, expr: N.Expression): N.ExpressionStatement;
    parseBlock(allowDirectives?: boolean, isFunctionScope?: boolean, contextId?: number): N.BlockStatement;
    isValidDirective(stmt: N.Statement): boolean;
    parseBlockBody(node: N.BlockStatementLike, allowDirectives: boolean, topLevel: boolean, end: TokenType): void;
    parseBlockOrModuleBlockBody(body: Array<N.Statement>, directives: Array<N.Directive> | null, topLevel: boolean, end: TokenType): void;
    parseFor(node: N.ForStatement, init: N.VariableDeclaration | N.Expression | null): N.ForStatement;
    parseForIn(node: N.ForInOf, init: N.VariableDeclaration, forAwait: boolean): N.ForInOf;
    parseVar(node: N.VariableDeclaration, isFor: boolean, kind: TokenType): N.VariableDeclaration;
    parseVarHead(decl: N.VariableDeclarator, isBlockScope: boolean): void;
    parseFunction<T extends N.NormalFunction>(node: T, isStatement: boolean, allowExpressionBody?: boolean, isAsync?: boolean, optionalId?: boolean): T;
    parseFunctionParams(node: N.Function, allowModifiers?: boolean, funcContextId?: number): void;
    parseClass<T extends N.Class>(node: T, isStatement: boolean, optionalId?: boolean): T;
    isClassProperty(): boolean;
    isClassMethod(): boolean;
    isNonstaticConstructor(method: N.ClassMethod | N.ClassProperty): boolean;
    parseClassBody(node: N.Class, classContextId: number): void;
    parseClassMember(classBody: N.ClassBody, member: N.ClassMember, state: {
        hadConstructor: boolean;
    }, classContextId: number): void;
    parseClassMemberWithIsStatic(classBody: N.ClassBody, member: N.ClassMember, state: {
        hadConstructor: boolean;
    }, isStatic: boolean, classContextId: number): void;
    parseClassPropertyName(member: N.ClassMember, classContextId: number): N.Expression | N.Identifier;
    pushClassProperty(classBody: N.ClassBody, prop: N.ClassProperty): void;
    pushClassPrivateProperty(classBody: N.ClassBody, prop: N.ClassPrivateProperty): void;
    pushClassMethod(classBody: N.ClassBody, method: N.ClassMethod, isGenerator: boolean, isAsync: boolean, isConstructor: boolean): void;
    pushClassPrivateMethod(classBody: N.ClassBody, method: N.ClassPrivateMethod, isGenerator: boolean, isAsync: boolean): void;
    parsePostMemberNameModifiers(methodOrProp: N.ClassMethod | N.ClassProperty): void;
    parseAccessModifier(): N.Accessibility | null;
    parseClassPrivateProperty(node: N.ClassPrivateProperty): N.ClassPrivateProperty;
    parseClassProperty(node: N.ClassProperty): N.ClassProperty;
    parseClassId(node: N.Class, isStatement: boolean, optionalId?: boolean): void;
    parseClassSuper(node: N.Class): void;
    parseExport(node: N.Node): N.Node;
    parseExportDefaultExpression(): N.Expression | N.Declaration;
    parseExportDeclaration(node: N.ExportNamedDeclaration): N.Declaration | null;
    isExportDefaultSpecifier(): boolean;
    parseExportSpecifiersMaybe(node: N.ExportNamedDeclaration): void;
    parseExportFrom(node: N.ExportNamedDeclaration, expect?: boolean): void;
    shouldParseExportStar(): boolean;
    parseExportStar(node: N.ExportNamedDeclaration): void;
    parseExportNamespace(node: N.ExportNamedDeclaration): void;
    shouldParseExportDeclaration(): boolean;
    checkExport(node: N.ExportNamedDeclaration, checkNames?: boolean, isDefault?: boolean): void;
    checkDeclaration(node: N.Pattern | N.ObjectProperty): void;
    checkDuplicateExports(node: N.Identifier | N.ExportNamedDeclaration | N.ExportSpecifier, name: string): void;
    raiseDuplicateExportError(node: N.Identifier | N.ExportNamedDeclaration | N.ExportSpecifier, name: string): never;
    parseExportSpecifiers(): Array<N.ExportSpecifier>;
    parseImport(node: N.Node): N.ImportDeclaration | N.TsImportEqualsDeclaration;
    shouldParseDefaultImport(node: N.ImportDeclaration): boolean;
    parseImportSpecifierLocal(node: N.ImportDeclaration, specifier: N.Node, type: string, contextDescription: string): void;
    parseImportSpecifiers(node: N.ImportDeclaration): void;
    parseImportSpecifier(node: N.ImportDeclaration): void;
}
