import { Options } from "../options";
import * as N from "../types";
import { Position } from "../util/location";
import { TokContext } from "./context";
import { Token } from "./index";
import { TokenType } from "./types";
export declare type Label = {
    name?: string;
    kind: "loop" | "switch" | null;
    statementStart?: number;
};
export declare type Scope = {
    isFunctionScope: boolean;
    startTokenIndex: number;
    endTokenIndex: number;
};
export default class State {
    init(options: Options, input: string): void;
    strict: boolean;
    input: string;
    potentialArrowAt: number;
    noArrowAt: Array<number>;
    noArrowParamsConversionAt: Array<number>;
    inFunction: boolean;
    inParameters: boolean;
    maybeInArrowParameters: boolean;
    inGenerator: boolean;
    inMethod: boolean | N.MethodKind;
    inAsync: boolean;
    inType: boolean;
    noAnonFunctionType: boolean;
    inPropertyName: boolean;
    inClassProperty: boolean;
    classLevel: number;
    labels: Array<Label>;
    decoratorStack: Array<Array<N.Decorator>>;
    yieldInPossibleArrowParameters: N.YieldExpression | null;
    tokens: Array<Token>;
    scopes: Array<Scope>;
    pos: number;
    lineStart: number;
    curLine: number;
    type: TokenType;
    value: any;
    start: number;
    end: number;
    isType: boolean;
    startLoc: Position;
    endLoc: Position;
    lastTokEndLoc: Position;
    lastTokStartLoc: Position;
    lastTokStart: number;
    lastTokEnd: number;
    context: Array<TokContext>;
    exprAllowed: boolean;
    containsEsc: boolean;
    containsOctal: boolean;
    octalPosition: number | null;
    exportedIdentifiers: Array<string>;
    invalidTemplateEscapePosition: number | null;
    curPosition(): Position;
    clone(skipArrays?: boolean): State;
}
