"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const location_1 = require("../util/location");
const context_1 = require("./context");
const types_1 = require("./types");
class State {
    init(options, input) {
        this.strict = options.strictMode === false ? false : options.sourceType === "module";
        this.input = input;
        this.potentialArrowAt = -1;
        this.noArrowAt = [];
        this.noArrowParamsConversionAt = [];
        this.inMethod = false;
        this.inFunction = false;
        this.inParameters = false;
        this.maybeInArrowParameters = false;
        this.inGenerator = false;
        this.inAsync = false;
        this.inPropertyName = false;
        this.inType = false;
        this.inClassProperty = false;
        this.noAnonFunctionType = false;
        this.classLevel = 0;
        this.labels = [];
        this.decoratorStack = [[]];
        this.yieldInPossibleArrowParameters = null;
        this.tokens = [];
        this.scopes = [];
        this.pos = 0;
        this.lineStart = 0;
        this.curLine = options.startLine;
        this.type = types_1.types.eof;
        this.value = null;
        this.start = this.pos;
        this.end = this.pos;
        this.startLoc = this.curPosition();
        this.endLoc = this.startLoc;
        this.isType = false;
        // @ts-ignore
        this.lastTokEndLoc = null;
        // @ts-ignore
        this.lastTokStartLoc = null;
        this.lastTokStart = this.pos;
        this.lastTokEnd = this.pos;
        this.context = [context_1.types.braceStatement];
        this.exprAllowed = true;
        this.containsEsc = false;
        this.containsOctal = false;
        this.octalPosition = null;
        this.invalidTemplateEscapePosition = null;
        this.exportedIdentifiers = [];
    }
    curPosition() {
        return new location_1.Position(this.curLine, this.pos - this.lineStart);
    }
    clone(skipArrays) {
        const state = new State();
        Object.keys(this).forEach((key) => {
            // $FlowIgnore
            let val = this[key];
            if ((!skipArrays || key === "context") && Array.isArray(val)) {
                val = val.slice();
            }
            // $FlowIgnore
            state[key] = val;
        });
        return state;
    }
}
exports.default = State;
